@@ -12,10 +12,14 @@ use crate::dialog::Dialog;
 use crate::errors::{Result, Unsupported};
 use crate::{Connection, XId};
 
-// Hide u32 because CompletionNotify events might not come in in order or the serial might have
-// wrapped.
 #[derive(Eq, PartialEq, Debug, Clone, Copy)]
-pub struct FrameId(u32);
+enum State {
+    Sync,
+    // Window needs redraw
+    Exposed,
+    // Windows needs redraw from our backbuffer
+    Dirty,
+}
 
 pub struct Backbuffer<'a> {
     conn: &'a Connection,
@@ -23,12 +27,13 @@ pub struct Backbuffer<'a> {
     eid: Option<XId>,
     serial: u32,
     vsync_completed: bool,
-    dirty: bool,
-    pub exposed: bool,
+    dirty: State,
     backbuffer_idle: bool,
     surface: XcbSurface<'a>,
     pub(super) cr: cairo::Context,
     pub(super) resize_requested: Option<(u16, u16)>,
+    // TODO how to know when the window is not exposed at all?
+    pub(super) visible: bool,
 }
 
 pub struct Cookie<'a> {
@@ -46,11 +51,12 @@ impl<'a> Cookie<'a> {
             let version = self.version.reply()?;
             let caps = self.caps.unwrap().reply()?;
             debug!(
-                "present version: {}.{}, capabilities: async {}, fence: {}",
+                "present version: {}.{}, capabilities: async {}, fence: {}, ust: {}",
                 version.major_version,
                 version.minor_version,
                 caps.capabilities & u32::from(present::Capability::ASYNC) != 0,
                 caps.capabilities & u32::from(present::Capability::FENCE) != 0,
+                caps.capabilities & u32::from(present::Capability::UST) != 0,
             );
         }
 
@@ -60,12 +66,12 @@ impl<'a> Cookie<'a> {
             eid: None,
             serial: 0,
             vsync_completed: true,
-            dirty: false,
+            dirty: State::Sync,
             backbuffer_idle: true,
             surface: self.surface,
             cr: self.cr,
             resize_requested: None,
-            exposed: false,
+            visible: false,
         };
         Ok(me)
     }
@@ -102,7 +108,16 @@ impl<'a> Backbuffer<'a> {
         })
     }
 
+    pub fn set_exposed(&mut self) {
+        trace!("set_exposed");
+        self.visible = true;
+        if self.vsync_completed && self.dirty == State::Sync {
+            self.dirty = State::Exposed;
+        }
+    }
+
     pub fn init(&mut self, window: xproto::Window, dialog: &mut Dialog) -> Result<()> {
+        trace!("init");
         self.eid = Some(self.conn.generate_id()?);
         self.conn.present_select_input(
             self.eid.unwrap(),
@@ -116,24 +131,29 @@ impl<'a> Backbuffer<'a> {
         self.surface.setup_pixmap(window, w, h)?;
         dialog.init(&self.cr);
         dialog.set_painted();
-        self.dirty = true;
+        self.dirty = State::Dirty;
         Ok(())
     }
 
     pub fn commit(&mut self, dialog: &mut Dialog) -> Result<()> {
+        trace!("commit");
+        if !self.visible {
+            debug!("not visible");
+            return Ok(());
+        }
         if dialog.dirty() || self.resize_requested.is_some() {
             self.repaint(dialog)?;
         }
-        if self.dirty || self.exposed {
+        if self.dirty != State::Sync {
             self.present(dialog)?;
         }
         Ok(())
     }
 
     fn repaint(&mut self, dialog: &mut Dialog) -> Result<()> {
+        trace!("repaint");
         if self.backbuffer_idle {
-            trace!("update");
-            self.dirty = true;
+            self.dirty = State::Dirty;
             if let Some((width, height)) = self.resize_requested {
                 trace!("resize requested");
                 let surface_cleared = self.surface.resize(width, height)?;
@@ -145,12 +165,13 @@ impl<'a> Backbuffer<'a> {
             self.surface.flush();
             dialog.set_painted();
         } else {
-            trace!("update: backbuffer not idle");
+            trace!("repaint: backbuffer not idle");
         }
         Ok(())
     }
 
     pub fn on_idle_notify(&mut self, ev: &present::IdleNotifyEvent) {
+        trace!("on_idle_notify: {:?}", ev);
         if ev.serial == self.serial {
             self.backbuffer_idle = true;
             trace!("idle notify: backbuffer became idle");
@@ -159,23 +180,18 @@ impl<'a> Backbuffer<'a> {
         }
     }
 
-    pub fn on_vsync_completed(&mut self, ev: present::CompleteNotifyEvent) -> FrameId {
-        trace!(
-            "on_vsync_completed: serial {}, msc {}, ust {}",
-            ev.serial,
-            ev.msc,
-            ev.ust
-        );
+    pub fn on_vsync_completed(&mut self, ev: present::CompleteNotifyEvent) {
+        trace!("on_vsync_completed: {:?}", ev);
         if ev.serial == self.serial {
-            assert_ne!(ev.mode, present::CompleteMode::SKIP);
+            assert_ne!(ev.mode, present::CompleteMode::SKIP, "Should not happen");
             self.vsync_completed = true;
         } else {
             panic!("on_vsync_completed: ev.serial != self.serial");
         }
-        FrameId(ev.serial)
     }
 
     fn present(&mut self, dialog: &mut Dialog) -> Result<()> {
+        trace!("present");
         if !self.vsync_completed {
             debug!(
                 "a frame (serial {}) already pending for present",
@@ -183,7 +199,6 @@ impl<'a> Backbuffer<'a> {
             );
             return Ok(());
         }
-        trace!("present");
         self.serial = self.get_next_serial();
         self.conn.present_pixmap(
             self.window,
@@ -196,15 +211,14 @@ impl<'a> Backbuffer<'a> {
             0,                            // target_crtc
             0,                            // wait_fence
             0,                            // idle_fence
-            present::Option::COPY.into(), // options
+            present::Option::NONE.into(), // options
             0,                            // target_msc
             0,   // divisor, if 0, the presentation occus after the current field
             0,   // remainder
             &[], // notifiers
         )?;
         self.backbuffer_idle = false;
-        self.dirty = false;
-        self.exposed = false;
+        self.dirty = State::Sync;
         self.vsync_completed = false;
 
         self.conn.flush()?;
@@ -221,7 +235,7 @@ impl<'a> Drop for Backbuffer<'a> {
     fn drop(&mut self) {
         debug!("dropping backbuffer");
         if let Some(eid) = self.eid {
-            if let Err(err) = self.conn.present_select_input(eid, self.window, 0u32) {
+            if let Err(err) = self.conn.present_select_input(eid, self.window, 0_u32) {
                 debug!("present select event clear failed: {}", err);
             }
         }
@@ -274,10 +288,13 @@ impl<'a> XcbSurface<'a> {
         height: u16,
     ) -> cairo::XCBSurface {
         let cairo_conn =
-            unsafe { cairo::XCBConnection::from_raw_none(conn.get_raw_xcb_connection() as _) };
+            unsafe { cairo::XCBConnection::from_raw_none(conn.get_raw_xcb_connection().cast()) };
         let mut xcb_visualtype: xcb_visualtype_t = (*visual_type).into();
-        let cairo_visual =
-            unsafe { cairo::XCBVisualType::from_raw_none(&mut xcb_visualtype as *mut _ as _) };
+        let cairo_visual = unsafe {
+            cairo::XCBVisualType::from_raw_none(
+                (&mut xcb_visualtype as *mut xcb_visualtype_t).cast(),
+            )
+        };
         let cairo_drawable = cairo::XCBDrawable(drawable);
         trace!("creating cairo::XCBSurface {}, {}", width, height);
         cairo::XCBSurface::create(
