@@ -1,10 +1,11 @@
 use std::cmp::{max, min};
-use std::convert::TryFrom as _;
+use std::convert::{TryFrom as _, TryInto as _};
 use std::ops::{Deref, DerefMut};
 use std::pin::Pin;
 use std::time::Duration;
 
 use log::{debug, log_enabled, trace};
+use pango::glib::translate::{FromGlib as _, ToGlibPtr as _};
 use rand::seq::SliceRandom as _;
 use tokio::time::{sleep, Instant, Sleep};
 
@@ -13,7 +14,37 @@ use crate::config;
 use crate::errors::Result;
 use crate::secret::{Passphrase, SecBuf};
 
+mod ffi {
+    #![allow(non_upper_case_globals)]
+    #![allow(non_camel_case_types)]
+    #![allow(non_snake_case)]
+    #![allow(dead_code)]
+    #![allow(deref_nullptr)]
+    #![allow(clippy::all, clippy::pedantic)]
+
+    include!(concat!(
+        env!("XASKPASS_BUILD_HEADER_DIR"),
+        "/pango_sys_fixes.rs"
+    ));
+}
+
+pub enum Direction {
+    Left,
+    Right,
+}
+
+impl From<Direction> for i32 {
+    fn from(dir: Direction) -> i32 {
+        if let Direction::Left = dir {
+            -1
+        } else {
+            1
+        }
+    }
+}
+
 #[derive(Debug)]
+#[allow(clippy::struct_excessive_bools)]
 pub struct Base {
     pub(super) x: f64,
     pub(super) y: f64,
@@ -125,10 +156,10 @@ impl Base {
     pub async fn handle_events(&mut self) {
         tokio::select! {
             _ = &mut self.blink_timeout, if self.blink_do => {
-                self.on_blink_timeout()
+                self.on_blink_timeout();
             }
             _ = &mut self.show_selection_timeout, if self.show_selection_do => {
-                self.on_show_selection_timeout()
+                self.on_show_selection_timeout();
             }
         }
     }
@@ -153,6 +184,7 @@ impl Base {
         cr.restore().unwrap();
     }
 
+    #[allow(clippy::too_many_arguments)]
     fn blink(
         &self,
         cr: &cairo::Context,
@@ -161,6 +193,7 @@ impl Base {
         y: f64,
         bg: Option<&Pattern>,
         sharp: bool,
+        width: f64,
     ) {
         cr.save().unwrap();
 
@@ -174,7 +207,7 @@ impl Base {
                 cr.move_to(x, y);
             };
             cr.rel_line_to(0.0, height);
-            cr.set_line_width(1.0);
+            cr.set_line_width(width);
             cr.stroke().unwrap();
         } else {
             cr.rectangle(x - 1.0, y - 1.0, 3.0, height + 2.0);
@@ -285,7 +318,9 @@ impl Circle {
         text_height: f64,
         debug: bool,
     ) -> Self {
-        let diameter = circle.diameter.unwrap_or(text_height * 3.0);
+        let diameter = circle
+            .diameter
+            .unwrap_or_else(|| (text_height * 3.0).round());
         let inner_radius =
             (diameter / 2.0 - circle.indicator_width.unwrap_or(diameter / 4.0)).max(0.0);
         let diameter = diameter + config.border_width * 2.0;
@@ -299,7 +334,7 @@ impl Circle {
         let indicator_count = circle.indicator_count;
         let spacing_angle = circle
             .spacing_angle
-            .min(2.0 * std::f64::consts::PI / indicator_count as f64);
+            .min(2.0 * std::f64::consts::PI / f64::from(indicator_count));
         let frame_increment_start = circle.rotation_speed_start;
         Self {
             base,
@@ -313,7 +348,7 @@ impl Circle {
             frame_increment: frame_increment_start,
             frame_increment_start,
             frame_increment_gain: circle.rotation_speed_gain,
-            angle: 2.0 * std::f64::consts::PI / indicator_count as f64,
+            angle: 2.0 * std::f64::consts::PI / f64::from(indicator_count),
             animation_distance: 0.0,
             rotation: 0.0,
             oldlen: 0,
@@ -348,11 +383,12 @@ impl Circle {
     }
 
     fn init_rotation(&mut self) {
-        trace!("run animation");
         const FULL_ROUND: f64 = 2.0 * std::f64::consts::PI;
+        trace!("run animation");
         self.rotation %= FULL_ROUND;
-        self.animation_distance += (self.pass.len as i64 - self.oldlen as i64) as f64
-            * (self.angle / self.indicator_count as f64);
+        self.animation_distance +=
+            f64::from(i32::try_from(self.pass.len).unwrap() - i32::try_from(self.oldlen).unwrap())
+                * (self.angle / f64::from(self.indicator_count));
         self.oldlen = self.pass.len;
         if self.animation_distance.abs() > 2.0 * FULL_ROUND {
             self.animation_distance %= FULL_ROUND;
@@ -422,13 +458,14 @@ impl Circle {
             (self.height - height) / 2.0,
             Some(&self.lock_color),
             false,
+            1.0,
         );
     }
 
     pub fn set_painted(&mut self) {
         trace!("set_painted paint_pending {:?}", self.paint_pending,);
         self.paint_pending = false;
-        self.base.set_painted()
+        self.base.set_painted();
     }
 
     // TODO
@@ -505,16 +542,17 @@ impl Circle {
             let is_lid = self.light_up
                 && self.pass.len > 0
                 && (self.show_selection_do
-                    || (i64::try_from(self.pass.len).unwrap() - 1) % self.indicator_count as i64
-                        == if self.rotate {
+                    || (i64::try_from(self.pass.len).unwrap() - 1)
+                        % i64::from(self.indicator_count)
+                        == i64::from(if self.rotate {
                             self.indicator_count - 1 - ix
                         } else {
                             ix
-                        } as i64);
+                        }));
 
             let rotation = self.rotation % (2.0 * std::f64::consts::PI);
-            let from_angle = self.angle * (ix as f64 - 1.0) + rotation;
-            let to_angle = self.angle * ix as f64 - self.spacing_angle + rotation;
+            let from_angle = self.angle * (f64::from(ix) - 1.0) + rotation;
+            let to_angle = self.angle * f64::from(ix) - self.spacing_angle + rotation;
 
             cr.new_path();
             cr.arc(middle.0, middle.1, stroke_radius, from_angle, to_angle);
@@ -550,7 +588,7 @@ impl Circle {
 }
 
 #[derive(Debug)]
-pub struct IndicatorElement {
+struct Element {
     x: f64,
     y: f64,
 }
@@ -565,7 +603,7 @@ pub struct Classic {
     horizontal_spacing: f64,
     radius_x: f64,
     radius_y: f64,
-    indicators: Vec<IndicatorElement>,
+    indicators: Vec<Element>,
     base: Base,
 }
 
@@ -593,7 +631,7 @@ impl Classic {
         let border_width = config.border_width;
         let element_height = classic
             .element_height
-            .unwrap_or(text_height + 2.0 * border_width);
+            .unwrap_or(text_height.ceil() + 2.0 * border_width);
         let height = element_height;
         let base = Base {
             height,
@@ -609,7 +647,7 @@ impl Classic {
             min_count: classic.min_count,
             element_width: classic
                 .element_width
-                .unwrap_or(text_height * 2.0 + 2.0 * border_width),
+                .unwrap_or(text_height.round() * 2.0 + 2.0 * border_width),
             element_height,
             radius_x: classic.radius_x,
             radius_y: classic.radius_y,
@@ -634,12 +672,12 @@ impl Classic {
             ),
             self.max_count,
         );
-        self.width = indicator_count as f64 * (self.element_width + self.horizontal_spacing)
+        self.width = f64::from(indicator_count) * (self.element_width + self.horizontal_spacing)
             - self.horizontal_spacing;
 
         let mut x = 0.0;
         for _ in 0..indicator_count {
-            let e = IndicatorElement { x, y: 0.0 };
+            let e = Element { x, y: 0.0 };
             self.indicators.push(e);
             x += self.element_width + self.horizontal_spacing;
         }
@@ -704,8 +742,7 @@ enum StringType {
 impl StringType {
     pub fn use_cursor(&self) -> bool {
         match self {
-            Self::Disco(..) => false,
-            Self::Custom(..) => false,
+            Self::Disco(..) | Self::Custom(..) => false,
             Self::Asterisk(..) => true,
         }
     }
@@ -766,8 +803,8 @@ impl Strings {
         strings_cfg: config::IndicatorStrings,
         layout: pango::Layout,
         debug: bool,
-        text_height: u32,
-    ) -> Result<Self> {
+        text_height: f64,
+    ) -> Self {
         let strings = match strings_cfg.strings {
             config::StringType::Asterisk { asterisk } => {
                 StringType::Asterisk(Asterisk::new(asterisk, &layout))
@@ -779,26 +816,26 @@ impl Strings {
         };
         let vertical_spacing = strings_cfg
             .vertical_spacing
-            .unwrap_or(text_height as f64 / 3.0)
+            .unwrap_or(text_height / 3.0)
             .round();
         let horizontal_spacing = strings_cfg
             .horizontal_spacing
-            .unwrap_or(text_height as f64 / 2.0)
+            .unwrap_or(text_height / 2.0)
             .round();
         debug!(
             "strings indicator: vertical_spacing: {}, horizontal_spacing: {}, border_width: {}",
             vertical_spacing, horizontal_spacing, config.border_width
         );
-        let height = text_height as f64 + 2.0 * vertical_spacing + 2.0 * config.border_width;
+        let height = text_height.ceil() + 2.0 * vertical_spacing + 2.0 * config.border_width;
         let base = Base {
             ..Base::new(config, height, debug)
         };
 
-        layout.set_height(i32::try_from(text_height).unwrap() * pango::SCALE);
+        layout.set_height((text_height * f64::from(pango::SCALE)).ceil() as i32);
         layout.set_single_paragraph_mode(true);
 
         let blink_spacing = if strings.use_cursor() { 0.0 } else { 8.0 };
-        Ok(Self {
+        Self {
             base,
             strings,
             radius_x: strings_cfg.radius_x,
@@ -811,7 +848,7 @@ impl Strings {
             show_plain: false,
             cursor: 0,
             hover: false,
-        })
+        }
     }
 
     pub fn is_inside(&mut self, x: f64, y: f64) -> bool {
@@ -859,32 +896,127 @@ impl Strings {
         }
     }
 
-    pub fn pass_delete(&mut self) {
-        trace!("pass delete {}", self.cursor);
-        self.base.key_pressed();
-        if self.cursor == 0 {
-            return;
+    fn get_log_attrs(layout: &pango::Layout) -> &[ffi::PangoLogAttr] {
+        unsafe {
+            let mut n_attrs: libc::c_int = 0;
+            let log_attrs = pango_sys::pango_layout_get_log_attrs_readonly(
+                layout.to_glib_none().0,
+                &mut n_attrs as *mut _,
+            );
+            assert!(!log_attrs.is_null());
+            std::slice::from_raw_parts(log_attrs.cast(), n_attrs.try_into().expect("n_attrs"))
         }
-        let i = self.cursor - 1;
-        self.pass.delete(i);
-        self.set_text();
-        self.cursor -= 1;
-        self.dirty = true;
     }
 
-    pub fn move_cursor(&mut self, direction: i32) {
+    fn move_backward_word(&self) -> usize {
+        if self.cursor == 0 || (!self.strings.use_cursor() && !self.show_plain) {
+            return 0;
+        }
+        let log_attrs = Self::get_log_attrs(&self.layout);
+        let mut cursor = self.cursor - 1;
+        while cursor > 0 && log_attrs[cursor].is_word_start() == 0 {
+            cursor -= 1;
+        }
+        cursor
+    }
+
+    fn move_forward_word(&self) -> usize {
+        if self.cursor >= self.pass.len || (!self.strings.use_cursor() && !self.show_plain) {
+            return self.pass.len;
+        }
+        let log_attrs = Self::get_log_attrs(&self.layout);
+        let mut cursor = self.cursor + 1;
+        while cursor < log_attrs.len() - 1 && log_attrs[cursor].is_word_end() == 0 {
+            cursor += 1;
+        }
+        cursor
+    }
+
+    fn move_backspace(&self) -> usize {
+        if self.cursor == 0 {
+            return 0;
+        }
+        let log_attrs = Self::get_log_attrs(&self.layout);
+        debug!("log_attrs len: {}", log_attrs.len());
+        if log_attrs[self.cursor].backspace_deletes_character() == 1 {
+            debug!(
+                "cursor {}, log_attrs: {:?}",
+                self.cursor, log_attrs[self.cursor]
+            );
+            self.cursor - 1
+        } else {
+            let mut new_cursor = self.cursor - 1;
+            while new_cursor > 0 && log_attrs[new_cursor].is_cursor_position() == 0 {
+                debug!("not a cursor position");
+                new_cursor -= 1;
+                debug!(
+                    "new_cursor: {}, log_attrs: {:?}",
+                    new_cursor, log_attrs[new_cursor]
+                );
+            }
+            new_cursor
+        }
+    }
+
+    pub fn pass_delete(&mut self, word: bool) {
+        trace!("pass delete {}", self.cursor);
+        self.base.key_pressed();
+        let new_cursor = if word {
+            self.move_backward_word()
+        } else {
+            self.move_backspace()
+        };
+        if new_cursor == self.cursor {
+            return;
+        }
+        assert!(new_cursor < self.cursor);
+        let old_index = self.cursor - 1;
+        self.pass.delete(new_cursor, old_index);
+        self.cursor = new_cursor;
+        self.dirty = true;
+        self.set_text();
+    }
+
+    pub fn move_visually(&mut self, direction: Direction, word: bool) {
         if !self.strings.use_cursor() && !self.show_plain {
             return;
         }
-        trace!("move cursor {}", self.cursor);
         self.key_pressed();
-        let new_cursor = self
-            .layout
-            .move_cursor_visually(true, self.cursor_bytes(), 0, direction);
-        if new_cursor.0 != std::i32::MAX && new_cursor.0 != -1 {
-            self.cursor = self.cursor_chars(new_cursor.0, new_cursor.1);
-
+        let new_cursor = if word {
+            let line = self.layout.line_readonly(0).unwrap();
+            let fixed_line: *const ffi::PangoLayoutLine = line.to_glib_none().0.cast();
+            let text_dir = unsafe {
+                pango::Direction::from_glib((*fixed_line).resolved_dir().try_into().unwrap())
+            };
+            debug!("text_direction: {:?}", text_dir);
+            if text_dir == pango::Direction::Rtl {
+                match direction {
+                    Direction::Right => self.move_backward_word(),
+                    Direction::Left => self.move_forward_word(),
+                }
+            } else {
+                match direction {
+                    Direction::Left => self.move_backward_word(),
+                    Direction::Right => self.move_forward_word(),
+                }
+            }
+        } else {
+            let new_cursor = self.layout.move_cursor_visually(
+                true,
+                self.cursor_bytes(self.cursor),
+                0,
+                direction.into(),
+            );
+            if new_cursor.0 != std::i32::MAX && new_cursor.0 != -1 {
+                self.cursor_chars(new_cursor.0, new_cursor.1)
+            } else {
+                return;
+            }
+        };
+        debug!("move cursor {} -> {}", self.cursor, new_cursor);
+        if new_cursor != self.cursor {
             self.dirty = true;
+            self.cursor = new_cursor;
         }
     }
 
@@ -901,19 +1033,19 @@ impl Strings {
         f.0 + usize::try_from(trailing).unwrap()
     }
 
-    fn cursor_bytes(&self) -> i32 {
+    fn cursor_bytes(&self, cursor: usize) -> i32 {
         assert!(self.strings.use_cursor() || self.show_plain);
-        if self.cursor == 0 {
+        if cursor == 0 {
             return 0;
         }
         let gs = self.layout.text().unwrap();
         let s = gs.as_str();
-        let indice = s.char_indices().nth(self.cursor - 1).unwrap();
+        let indice = s.char_indices().nth(cursor - 1).unwrap();
         i32::try_from(indice.0 + indice.1.len_utf8()).unwrap()
     }
 
     pub fn for_width(&mut self, for_width: f64) {
-        self.width = self.strings.for_width(&self.layout, for_width) as f64
+        self.width = f64::from(self.strings.for_width(&self.layout, for_width))
             + 2.0 * self.horizontal_spacing
             + self.blink_spacing
             + 2.0 * self.border_width;
@@ -926,6 +1058,11 @@ impl Strings {
         }
 
         self.set_text();
+
+        let log_attrs = Self::get_log_attrs(&self.layout);
+        while self.cursor < log_attrs.len() && log_attrs[self.cursor].is_cursor_position() == 0 {
+            self.cursor += 1;
+        }
     }
 
     pub fn into_pass(self) -> Passphrase {
@@ -986,10 +1123,10 @@ impl Strings {
     pub async fn handle_events(&mut self) {
         tokio::select! {
             _ = &mut self.base.blink_timeout, if self.base.blink_do => {
-                self.on_blink_timeout()
+                self.on_blink_timeout();
             }
             _ = &mut self.base.show_selection_timeout, if self.base.show_selection_do => {
-                self.on_show_selection_timeout()
+                self.on_show_selection_timeout();
             }
         }
     }
@@ -1021,7 +1158,7 @@ impl Strings {
                             - self.blink_spacing
                             - self.horizontal_spacing
                             - self.border_width)
-                            * pango::SCALE as f64) as i32,
+                            * f64::from(pango::SCALE)) as i32,
                         rec.x,
                     ),
                     rec.x + rec.width - 1,
@@ -1029,7 +1166,7 @@ impl Strings {
                 min(
                     max(
                         ((y - self.y - self.vertical_spacing - self.border_width)
-                            * pango::SCALE as f64) as i32,
+                            * f64::from(pango::SCALE)) as i32,
                         rec.y,
                     ),
                     rec.y + rec.height - 1,
@@ -1040,18 +1177,18 @@ impl Strings {
                 self.cursor = self.cursor_chars(idx, trailing);
                 self.dirty = true;
                 return true;
-            } else {
-                assert!(
-                    self.pass.len == 0,
-                    "click x:{}, y: {}, {} {} {}",
-                    x,
-                    y,
-                    inside,
-                    idx,
-                    trailing
-                );
-                return false;
             }
+
+            assert!(
+                self.pass.len == 0,
+                "click x:{}, y: {}, {} {} {}",
+                x,
+                y,
+                inside,
+                idx,
+                trailing
+            );
+            return false;
         }
         false
     }
@@ -1076,22 +1213,40 @@ impl Strings {
     fn blink(&self, cr: &cairo::Context) {
         if self.has_focus && self.blink_on {
             let pos = if self.show_plain || self.strings.use_cursor() {
-                (self.layout.cursor_pos(self.cursor_bytes()).0.x as f64 / pango::SCALE as f64)
-                    .round()
-                    + self.blink_spacing
+                let pos = self.layout.cursor_pos(self.cursor_bytes(self.cursor));
+                (pos.0.x, pos.1.x)
             } else {
-                0.0
+                (0, 0)
             };
             self.base.blink(
                 cr,
                 self.height - 2.0 * self.vertical_spacing - 2.0 * self.border_width,
-                self.border_width + self.horizontal_spacing + pos,
+                self.border_width
+                    + self.horizontal_spacing
+                    + (f64::from(pos.0) / f64::from(pango::SCALE)).round()
+                    + self.blink_spacing,
                 self.vertical_spacing + self.border_width,
                 None,
                 true,
-            )
+                1.0,
+            );
+            if pos.0 != pos.1 {
+                debug!("strong cursor: {}, weak cursor: {}", pos.0, pos.1);
+                self.base.blink(
+                    cr,
+                    self.height - 2.0 * self.vertical_spacing - 2.0 * self.border_width,
+                    self.border_width
+                        + self.horizontal_spacing
+                        + (f64::from(pos.1) / f64::from(pango::SCALE)).round()
+                        + self.blink_spacing,
+                    self.vertical_spacing + self.border_width,
+                    None,
+                    false,
+                    0.5,
+                );
+            }
         } else {
-            self.paint(cr)
+            self.paint(cr);
         }
     }
 }
@@ -1163,14 +1318,14 @@ impl Disco {
             })
             .collect();
         // every string with the same font should have the same logical height
-        let widths = sizes.iter().map(|(w, _)| *w as f64).collect();
-        let dancer_max_width = sizes.into_iter().map(|(w, _)| w).max().unwrap() as f64;
+        let widths = sizes.iter().map(|(w, _)| f64::from(*w)).collect();
+        let dancer_max_width = f64::from(sizes.into_iter().map(|(w, _)| w).max().unwrap());
         layout.set_text("");
         trace!("disco new end");
         Self {
             widths,
             dancer_max_width,
-            separator_width: layout.pixel_size().1 as f64,
+            separator_width: f64::from(layout.pixel_size().1),
             config,
             dancer_count: 0,
         }
@@ -1205,7 +1360,7 @@ impl Disco {
     }
 
     pub fn set_text(&mut self, layout: &pango::Layout, pass: &SecBuf<char>, show_paste: bool) {
-        self.set_text_do(layout, pass.len, show_paste)
+        self.set_text_do(layout, pass.len, show_paste);
     }
 
     fn set_text_do(&mut self, layout: &pango::Layout, pass_len: usize, show_paste: bool) {
@@ -1250,7 +1405,7 @@ impl Asterisk {
         layout.set_alignment(config.alignment.into());
         layout.set_text("");
         Self {
-            asterisk_width: asterisk_width as f64,
+            asterisk_width: f64::from(asterisk_width),
             asterisk,
             min_count: config.min_count,
             max_count: config.max_count,
