@@ -3,15 +3,16 @@ use tokio::io::unix::AsyncFd;
 use tokio::time::Instant;
 use x11rb::connection::Connection as _;
 use x11rb::connection::RequestConnection as _;
+use x11rb::cookie::Cookie;
 use x11rb::protocol::xfixes::{self, ConnectionExt as _};
 use x11rb::protocol::xproto::{self, ConnectionExt as _, CursorWrapper, WindowWrapper};
-use x11rb::protocol::Event as XEvent;
+use x11rb::protocol::Event;
 use zeroize::Zeroize;
 
 use crate::backbuffer::Backbuffer;
 use crate::dialog::{Action, Dialog};
 use crate::errors::{Error, Result, Unsupported};
-use crate::keyboard::{Keyboard, Keycode};
+use crate::keyboard::Keyboard;
 use crate::secret::Passphrase;
 use crate::Connection;
 
@@ -21,22 +22,13 @@ enum State {
     Cancelled,
 }
 
-pub struct Keypress {
-    key_press: xproto::KeyPressEvent,
-}
-
-impl Keypress {
-    pub fn get_key(&self) -> Keycode {
-        self.key_press.detail.into()
-    }
-}
-
-impl Drop for Keypress {
-    fn drop(&mut self) {
-        self.key_press.detail.zeroize();
-    }
+#[derive(Debug)]
+enum Reply {
+    GrabKeyboard(xproto::GrabKeyboardReply),
+    Selection(xproto::GetPropertyReply),
 }
 
+#[allow(clippy::struct_excessive_bools)]
 pub struct XContext<'a> {
     pub xfd: &'a AsyncFd<Connection>,
     pub backbuffer: Backbuffer<'a>,
@@ -48,13 +40,16 @@ pub struct XContext<'a> {
     pub(super) grab_keyboard: bool,
     pub(super) startup_time: Instant,
     pub(super) keyboard_grabbed: bool,
-    pub(super) first_expose_received: bool,
     pub(super) input_cursor: Option<CursorWrapper<'a, Connection>>,
     pub(super) compositor_atom: Option<xproto::Atom>,
+    pub(super) grab_keyboard_cookie: Option<Cookie<'a, Connection, xproto::GrabKeyboardReply>>,
+    pub(super) selection_cookie: Option<Cookie<'a, Connection, xproto::GetPropertyReply>>,
+    pub(super) debug: bool,
+    pub(super) first_expose_received: bool,
 }
 
 impl<'a> XContext<'a> {
-    pub fn conn(&self) -> &Connection {
+    pub fn conn(&self) -> &'a Connection {
         self.xfd.get_ref()
     }
 
@@ -84,10 +79,23 @@ impl<'a> XContext<'a> {
         Ok(())
     }
 
+    // TODO want cookie.poll_for_reply
+    fn poll_for_reply(&mut self) -> Result<Option<Reply>> {
+        Ok(if let Some(cookie) = self.selection_cookie.take() {
+            Some(Reply::Selection(cookie.reply()?))
+        } else if let Some(cookie) = self.grab_keyboard_cookie.take() {
+            Some(Reply::GrabKeyboard(cookie.reply()?))
+        } else {
+            None
+        })
+    }
+
     pub async fn run_events(&mut self, mut dialog: Dialog) -> Result<Option<Passphrase>> {
-        tokio::pin! { let xevents_ready = self.xfd.readable(); }
+        tokio::pin! { let events_ready = self.xfd.readable(); }
         dialog.init_events();
         loop {
+            // TODO do not draw if the window is not exposed at all
+            self.backbuffer.commit(&mut dialog)?;
             self.conn().flush()?;
             tokio::select! {
                 action = dialog.handle_events() => {
@@ -95,21 +103,30 @@ impl<'a> XContext<'a> {
                         return Ok(None)
                     }
                 }
-                xevents_guard = &mut xevents_ready => {
-                    if let Some(xevent) = self.conn().poll_for_event()? {
-                        //silly!("xevent {:?}", xevent);
-                        match self.handle_xevent(&mut dialog, xevent)? {
-                            State::Continue => {},
-                            State::Ready => { return Ok(Some(dialog.indicator.into_pass())) },
-                            State::Cancelled => { return Ok(None) },
+                events_guard = &mut events_ready => {
+                    let state = if let Some(event) = self.conn().poll_for_event()? {
+                        if self.debug {
+                            trace!("event {:?}", event);
                         }
+                        self.handle_event(&mut dialog, event)?
+                    } else if let Some(reply) = self.poll_for_reply()? {
+                        if self.debug {
+                            trace!("reply {:?}", reply);
+                        }
+                        self.handle_reply(&mut dialog, reply);
+                        State::Continue
                     } else {
-                        xevents_guard.unwrap().clear_ready();
+                        events_guard.unwrap().clear_ready();
+                        State::Continue
+                    };
+                    events_ready.set(self.xfd.readable());
+                    match state {
+                        State::Continue => {},
+                        State::Ready => { return Ok(Some(dialog.indicator.into_pass())) },
+                        State::Cancelled => { return Ok(None) },
                     }
-                    xevents_ready.set(self.xfd.readable());
                 }
             }
-            self.backbuffer.commit(&mut dialog)?;
             tokio::task::yield_now().await;
         }
     }
@@ -157,17 +174,18 @@ impl<'a> XContext<'a> {
         Ok(())
     }
 
-    fn handle_xevent(&mut self, dialog: &mut Dialog, event: XEvent) -> Result<State> {
+    #[allow(clippy::too_many_lines)]
+    fn handle_event(&mut self, dialog: &mut Dialog, event: Event) -> Result<State> {
         match event {
-            XEvent::Error(error) => {
+            Event::Error(error) => {
                 return Err(Error::X11(error));
             }
-            XEvent::Expose(expose_event) => {
+            Event::Expose(expose_event) => {
                 if expose_event.count > 0 {
                     return Ok(State::Continue);
                 }
 
-                self.backbuffer.exposed = true;
+                self.backbuffer.set_exposed();
 
                 if !self.first_expose_received {
                     debug!(
@@ -175,30 +193,23 @@ impl<'a> XContext<'a> {
                         self.startup_time.elapsed().as_millis()
                     );
                     self.first_expose_received = true;
+                }
 
-                    if self.grab_keyboard {
-                        let grabbed = self
-                            .conn()
-                            .grab_keyboard(
-                                false,
-                                self.window.window(),
-                                x11rb::CURRENT_TIME,
-                                xproto::GrabMode::ASYNC,
-                                xproto::GrabMode::ASYNC,
-                            )?
-                            .reply()?
-                            .status;
-                        if matches!(grabbed, xproto::GrabStatus::SUCCESS) {
-                            // TODO should set_focus(true) if focus event is not implied
-                            self.keyboard_grabbed = true;
-                            debug!("keyboard grab succeeded");
-                        } else {
-                            warn!("keyboard grab failed: {:?}", grabbed);
-                        }
+                if self.grab_keyboard && !self.keyboard_grabbed {
+                    if self.grab_keyboard_cookie.is_some() {
+                        debug!("grab keyboard already requested");
+                    } else {
+                        self.grab_keyboard_cookie = Some(self.conn().grab_keyboard(
+                            false,
+                            self.window.window(),
+                            x11rb::CURRENT_TIME,
+                            xproto::GrabMode::ASYNC,
+                            xproto::GrabMode::ASYNC,
+                        )?);
                     }
                 }
             }
-            XEvent::ConfigureNotify(ev) => {
+            Event::ConfigureNotify(ev) => {
                 if self.width != ev.width || self.height != ev.height {
                     trace!("resize event w: {}, h: {}", ev.width, ev.height);
                     self.width = ev.width;
@@ -206,17 +217,12 @@ impl<'a> XContext<'a> {
                     self.backbuffer.resize_requested = Some((ev.width, ev.height));
                 }
             }
-            // minimized
-            XEvent::UnmapNotify(..) => {}
-            // unminimized
-            XEvent::MapNotify(..) => {}
-            XEvent::ReparentNotify(..) => {}
-            XEvent::MotionNotify(me) => {
+            Event::MotionNotify(me) => {
                 if me.same_screen {
                     let (x, y) = self
                         .backbuffer
                         .cr
-                        .device_to_user(me.event_x as f64, me.event_y as f64)
+                        .device_to_user(f64::from(me.event_x), f64::from(me.event_y))
                         .expect("cairo device_to_user");
                     dialog.handle_motion(x, y, self)?;
                 } else {
@@ -224,8 +230,8 @@ impl<'a> XContext<'a> {
                 }
             }
             // both events have the same structure
-            XEvent::ButtonPress(bp) | XEvent::ButtonRelease(bp) => {
-                let isrelease = matches!(event, XEvent::ButtonRelease(_));
+            Event::ButtonPress(bp) | Event::ButtonRelease(bp) => {
+                let isrelease = matches!(event, Event::ButtonRelease(_));
                 trace!(
                     "button {}: {:?}",
                     if isrelease { "release" } else { "press" },
@@ -233,25 +239,23 @@ impl<'a> XContext<'a> {
                 );
                 if !bp.same_screen {
                     trace!("not same screen");
-                } else {
-                    let (x, y) = self
-                        .backbuffer
-                        .cr
-                        .device_to_user(bp.event_x as f64, bp.event_y as f64)
-                        .expect("cairo device_to_user");
-                    let action =
-                        dialog.handle_button_press(bp.detail.into(), x, y, isrelease, self)?;
-                    match action {
-                        Action::Ok => return Ok(State::Ready),
-                        Action::Cancel => return Ok(State::Cancelled),
-                        Action::Nothing => {}
-                        _ => unreachable!(),
-                    }
+                    return Ok(State::Continue);
+                }
+                let (x, y) = self
+                    .backbuffer
+                    .cr
+                    .device_to_user(f64::from(bp.event_x), f64::from(bp.event_y))
+                    .expect("cairo device_to_user");
+                let action = dialog.handle_button_press(bp.detail.into(), x, y, isrelease, self)?;
+                match action {
+                    Action::Ok => return Ok(State::Ready),
+                    Action::Cancel => return Ok(State::Cancelled),
+                    Action::Nothing => {}
+                    _ => unreachable!(),
                 }
             }
-            XEvent::KeyRelease(..) => {}
-            XEvent::KeyPress(key_press) => {
-                let action = dialog.handle_key_press(Keypress { key_press }, self)?;
+            Event::KeyPress(key_press) => {
+                let action = dialog.handle_key_press(key_press.detail.into(), self)?;
                 trace!("action {:?}", action);
                 match action {
                     Action::Ok => return Ok(State::Ready),
@@ -260,51 +264,41 @@ impl<'a> XContext<'a> {
                     _ => unreachable!(),
                 }
             }
-            XEvent::SelectionNotify(sn) => {
+            Event::SelectionNotify(sn) => {
                 if sn.property == x11rb::NONE {
-                    warn!("selection failed?");
-                } else {
-                    let reply = self
-                        .conn()
-                        .get_property(
-                            false,
-                            sn.requestor,
-                            sn.property,
-                            xproto::GetPropertyType::ANY,
-                            0,
-                            u32::MAX,
-                        )?
-                        .reply()?;
-                    if reply.format != 8 {
-                        warn!("invalid selection format {}", reply.format);
-                    // TODO
-                    } else if reply.type_ == self.atoms.INCR {
-                        warn!("Data too large and INCR mechanism not implemented");
-                    } else {
-                        match String::from_utf8(reply.value) {
-                            Err(err) => {
-                                warn!("selection not valid utf8: {}", err);
-                                err.into_bytes().zeroize();
-                            }
-                            Ok(mut val) => {
-                                dialog.indicator.pass_insert(&val, true);
-                                val.zeroize();
-                            }
-                        }
-                    }
+                    warn!("invalid selection");
+                    return Ok(State::Continue);
                 }
+                self.selection_cookie = Some(self.conn().get_property(
+                    false,
+                    sn.requestor,
+                    sn.property,
+                    xproto::GetPropertyType::ANY,
+                    0,
+                    u32::MAX,
+                )?);
             }
-            XEvent::FocusIn(..) => {
+            Event::FocusIn(fe) => {
+                if fe.mode == xproto::NotifyMode::GRAB {
+                    self.keyboard_grabbed = true;
+                } else if fe.mode == xproto::NotifyMode::UNGRAB {
+                    self.keyboard_grabbed = false;
+                }
                 dialog.indicator.set_focused(true);
             }
-            XEvent::FocusOut(fe) => {
+            Event::FocusOut(fe) => {
+                if fe.mode == xproto::NotifyMode::GRAB {
+                    self.keyboard_grabbed = true;
+                } else if fe.mode == xproto::NotifyMode::UNGRAB {
+                    self.keyboard_grabbed = false;
+                }
                 if fe.mode != xproto::NotifyMode::GRAB
                     && fe.mode != xproto::NotifyMode::WHILE_GRABBED
                 {
                     dialog.indicator.set_focused(false);
                 }
             }
-            XEvent::ClientMessage(client_message) => {
+            Event::ClientMessage(client_message) => {
                 debug!("client message");
                 if client_message.format == 32
                     && client_message.data.as_data32()[0] == self.atoms.WM_DELETE_WINDOW
@@ -313,35 +307,78 @@ impl<'a> XContext<'a> {
                     return Ok(State::Cancelled);
                 }
             }
-            XEvent::PresentIdleNotify(ev) => {
+            Event::PresentIdleNotify(ev) => {
                 self.backbuffer.on_idle_notify(&ev);
             }
-            XEvent::PresentCompleteNotify(ev) => {
+            Event::PresentCompleteNotify(ev) => {
                 self.backbuffer.on_vsync_completed(ev);
             }
-            XEvent::XkbStateNotify(key) => {
+            Event::XkbStateNotify(key) => {
                 self.keyboard.update_mask(&key);
             }
             // TODO needs more testing
-            XEvent::XkbNewKeyboardNotify(..) => {
+            Event::XkbNewKeyboardNotify(..) => {
                 debug!("xkb new keyboard notify");
                 self.keyboard.reload_keymap();
             }
             // TODO needs more testing
-            XEvent::XkbMapNotify(..) => {
+            Event::XkbMapNotify(..) => {
                 debug!("xkb map notify");
                 self.keyboard.reload_keymap();
             }
-            XEvent::XfixesSelectionNotify(sn) => {
+            Event::XfixesSelectionNotify(sn) => {
                 debug!("selection notify: {:?}", sn);
                 dialog.set_transparency(sn.subtype == xfixes::SelectionEvent::SET_SELECTION_OWNER);
             }
+            // minimized
+            Event::UnmapNotify(..) => {
+                debug!("set invisible");
+                self.backbuffer.visible = false;
+            }
+            // Ignored events:
+            // unminimized
+            Event::MapNotify(..) | Event::ReparentNotify(..) | Event::KeyRelease(..) => {
+                trace!("ignored event {:?}", event);
+            }
             event => {
                 debug!("unexpected event {:?}", event);
             }
         }
         Ok(State::Continue)
     }
+
+    fn handle_reply(&mut self, dialog: &mut Dialog, reply: Reply) {
+        match reply {
+            Reply::GrabKeyboard(gk) => {
+                let grabbed = gk.status;
+                match grabbed {
+                    xproto::GrabStatus::SUCCESS => debug!("keyboard grab succeeded"),
+                    xproto::GrabStatus::ALREADY_GRABBED => debug!("keyboard already grabbed"),
+                    _ => warn!("keyboard grab failed: {:?}", grabbed),
+                }
+            }
+            Reply::Selection(selection) => {
+                if selection.format != 8 {
+                    warn!("invalid selection format {}", selection.format);
+                    return;
+                // TODO
+                } else if selection.type_ == self.atoms.INCR {
+                    warn!("Selection too big and INCR selection not implemented");
+                    return;
+                }
+                match String::from_utf8(selection.value) {
+                    Err(err) => {
+                        warn!("selection is not valid utf8: {}", err);
+                        err.into_bytes().zeroize();
+                    }
+                    Ok(mut val) => {
+                        dialog.indicator.pass_insert(&val, true);
+                        val.zeroize();
+                    }
+                }
+            }
+        }
+    }
 }
 
 impl<'a> Drop for XContext<'a> {
@@ -356,7 +393,7 @@ impl<'a> Drop for XContext<'a> {
                 self.conn(),
                 self.window.window(),
                 compositor_atom,
-                0u32,
+                0_u32,
             ) {
                 debug!("clear select selection failed: {}", err);
             }
