@@ -3,7 +3,7 @@
 #![allow(non_snake_case)]
 #![allow(dead_code)]
 #![allow(deref_nullptr)]
-#![allow(clippy::all)]
+#![allow(clippy::all, clippy::pedantic)]
 
 include!(concat!(
     env!("XASKPASS_BUILD_HEADER_DIR"),
