@@ -1,3 +1,9 @@
+#![warn(clippy::all, clippy::pedantic)]
+#![allow(clippy::used_underscore_binding)]
+#![allow(clippy::non_ascii_literal)]
+#![allow(clippy::cast_sign_loss, clippy::cast_possible_truncation)]
+#![allow(clippy::option_if_let_else)]
+
 use std::convert::TryInto as _;
 use std::os::unix::ffi::OsStrExt as _;
 use std::path::{Path, PathBuf};
@@ -56,7 +62,7 @@ pub type XId = u32;
 
 pub type Connection = XCBConnection;
 
-/// Modified from https://github.com/psychon/x11rb/blob/master/cairo-example/src/main.rs
+/// Modified from <https://github.com/psychon/x11rb/blob/master/cairo-example/src/main.rs>
 /// Choose a visual to use. This function tries to find a depth=32 visual and falls back to the
 /// screen's default visual.
 fn choose_visual(conn: &Connection, screen_num: usize) -> Result<(u8, xproto::Visualid)> {
@@ -111,6 +117,7 @@ fn find_xcb_visualtype(conn: &Connection, visual_id: u32) -> Option<xproto::Visu
     None
 }
 
+#[allow(clippy::too_many_lines)]
 async fn run_xcontext(
     config: config::Config,
     opts: Opts,
@@ -122,13 +129,6 @@ async fn run_xcontext(
 
     debug!("connected X server");
     let atoms = AtomCollection::new(conn)?;
-    let compositor_atom = if config.depth == 32 {
-        conn.prefetch_extension_information(x11rb::protocol::xfixes::X11_EXTENSION_NAME)?;
-        let compositor_atom = format!("_NET_WM_CM_S{}", screen_num);
-        Some(conn.intern_atom(false, compositor_atom.as_bytes())?)
-    } else {
-        None
-    };
 
     conn.prefetch_extension_information(x11rb::protocol::present::X11_EXTENSION_NAME)?;
     conn.prefetch_extension_information(x11rb::protocol::xkb::X11_EXTENSION_NAME)?;
@@ -145,6 +145,15 @@ async fn run_xcontext(
         (screen.root_depth, screen.root_visual)
     };
     debug!("window depth: {}", depth);
+
+    let compositor_atom = if depth == 32 {
+        conn.prefetch_extension_information(x11rb::protocol::xfixes::X11_EXTENSION_NAME)?;
+        let compositor_atom = format!("_NET_WM_CM_S{}", screen_num);
+        Some(conn.intern_atom(false, compositor_atom.as_bytes())?)
+    } else {
+        None
+    };
+
     let visual_type = find_xcb_visualtype(conn, visualid).unwrap();
 
     let surface = backbuffer::XcbSurface::new(conn, screen.root, depth, &visual_type, 1, 1)?;
@@ -200,19 +209,14 @@ async fn run_xcontext(
             .colormap(
                 colormap
                     .as_ref()
-                    .map(|c| c.colormap())
-                    .unwrap_or(screen.default_colormap),
+                    .map_or(screen.default_colormap, ColormapWrapper::colormap),
             ),
     )?;
     let window = window_wrapper.window();
 
     let atoms = atoms.reply()?;
 
-    let hostname = if let Some(hn) = std::env::var_os("HOSTNAME") {
-        hn
-    } else {
-        gethostname::gethostname()
-    };
+    let hostname = std::env::var_os("HOSTNAME").unwrap_or_else(gethostname::gethostname);
     let mut title = config.title;
     if config.show_hostname {
         title.push('@');
@@ -351,7 +355,7 @@ async fn run_xcontext(
     };
 
     let resource_db;
-    let cursor_handle = if dialog.uses_cursor {
+    let cursor_handle = if dialog.uses_cursor() {
         resource_db = x11rb::resource_manager::Database::new_from_default(conn)?;
         Some(x11rb::cursor::Handle::new(conn, screen_num, &resource_db)?)
     } else {
@@ -361,12 +365,9 @@ async fn run_xcontext(
     debug!("compositor detected: {}", transparency);
     dialog.set_transparency(transparency);
 
-    debug!("dialog init");
-    let mut backbuffer = backbuffer.reply()?;
-    backbuffer.init(window, &mut dialog)?;
-
     debug!("keyboard init");
     let keyboard = keyboard::Keyboard::new(conn)?;
+    dialog.set_keyboard(&keyboard);
 
     debug!("cursor init");
     let input_cursor = if let Some(cursor_handle) = cursor_handle {
@@ -378,6 +379,11 @@ async fn run_xcontext(
     } else {
         None
     };
+
+    debug!("dialog init");
+    let mut backbuffer = backbuffer.reply()?;
+    backbuffer.init(window, &mut dialog)?;
+
     let mut xcontext = event::XContext {
         keyboard,
         xfd: &xfd,
@@ -388,10 +394,13 @@ async fn run_xcontext(
         height: window_height,
         grab_keyboard: config.grab_keyboard,
         startup_time,
-        first_expose_received: false,
         keyboard_grabbed: false,
         input_cursor,
         compositor_atom,
+        selection_cookie: None,
+        grab_keyboard_cookie: None,
+        debug: opts.debug,
+        first_expose_received: false,
     };
 
     xcontext.init()?;
@@ -425,7 +434,7 @@ struct Opts {
     config: Option<PathBuf>,
 
     #[clap(short, long)]
-    /// Include timestamps and sensitive information in logs.
+    /// Include additional and sensitive information in logs.
     debug: bool,
 
     /// Label in the dialog.
@@ -452,26 +461,31 @@ impl Opts {
 fn run() -> i32 {
     let startup_time = Instant::now();
 
-    let app = Opts::into_app();
     let cfg_loader = config::Loader::new();
-    let help = format!(
-        "CONFIGURATION FILE:\n    default: {}{}.toml",
+    let mut help = format!(
+        "CONFIGURATION FILE:\n    defaults: {}{}.toml",
         cfg_loader.xdg_dirs.get_config_home().display(),
         config::NAME,
     );
-    let app = app.after_help(&*help);
+    for d in cfg_loader.xdg_dirs.get_config_dirs() {
+        help.push_str(&format!(
+            ",\n              {}/{}.toml",
+            d.display(),
+            config::NAME
+        ));
+    }
+    let app = Opts::into_app().after_help(&*help);
     let opts = Opts::from_arg_matches(&app.get_matches()).expect("from_arg_matches");
 
     let mut log = stderrlog::new();
-    log.quiet(opts.quiet)
-        .verbosity(opts.verbose + 2)
-        .show_module_names(true);
+    log.quiet(opts.quiet).verbosity(opts.verbose + 2);
     if opts.debug {
-        log.timestamp(stderrlog::Timestamp::Millisecond);
+        log.timestamp(stderrlog::Timestamp::Millisecond)
+            .show_module_names(true);
     }
     log.init().unwrap();
 
-    match run_logged(cfg_loader, opts, startup_time) {
+    match run_logged(&cfg_loader, opts, startup_time) {
         Ok(ret) => ret,
         Err(err) => {
             error!("{}", err);
@@ -480,16 +494,16 @@ fn run() -> i32 {
     }
 }
 
-fn run_logged(cfg_loader: config::Loader, opts: Opts, startup_time: Instant) -> Result<i32> {
+fn run_logged(cfg_loader: &config::Loader, opts: Opts, startup_time: Instant) -> Result<i32> {
     if opts.gen_config {
         let cfg = config::Config::default();
-        cfg_loader.print(&cfg)?;
+        config::Loader::print(&cfg)?;
         return Ok(0);
     }
 
     debug!("load config");
     let config = if let Some(ref path) = opts.config {
-        cfg_loader.load_path(path)?
+        config::Loader::load_path(path)?
     } else {
         cfg_loader.load()?
     };
